/*****************************************************************************\
 *
 * perlin.cpp
 *
 * perlin - generate terrain and create chunked-LOD structure
 *
 * Authors: Petr "Lamik" Vrba (xvrbap00 _AT stud.fit.vutbr.cz)
 *
 * ----------------------------------------------------------------------------
 *
 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 * This source code is offered for use in the public domain.
 * You may use, modify or distribute it freely.
 *
 * This source code is distributed in the hope that it will be useful but
 * WITHOUT ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY
 * DISCLAIMED.  This includes but is not limited to warranties of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * If you find the source code useful, authors will kindly welcome
 * if you give them credit and keep their names with their source code,
 * but do not feel to be forced to do so.
 *
\*****************************************************************************/

#include "perlin.h"


#ifdef _MSC_VER
float strtof(const char *nptr, char **endptr)  { return float(strtod(nptr, endptr)); }
#endif

const char *appTitle = "Perlin function terrain generator and chunked-LOD renderer";      // Titulek okna
bool headlight = false;                           // Defaultn svtlo (v pozici kamery)
SoSeparator* createScene(int argc, char **argv);  // Funkce vytvejc scnu

//Vzdalenost od ktere bude klesat detail segmentu
float LODdistance = 55;

//globalni promenne pro minimalni a maximalni vysku
float minHeight =  999999;
float maxHeight = -999999;

//pocet jiz vygenerovanych segmentu
int already_finished = 0;
int total_segments = 0;


//Nainportuju kameru
extern SoCamera *kamera;

// Globalni promenne a funkce,
// ktere jsou pouzity v modulu main-windows.cpp
TerrainPatch *Terrain;
tmpTerrainPatch *tmpTerrain;

LODtable *myLODtable;

/*!
  Funkce alokuje pamet pro vyskovou mapu.
 */
void tmpTerrainPatch::allocHmap(int dimension)
{
  if (this->hmap == NULL || this->dimension != dimension) {
    delete[] this->hmap;
    this->dimension = dimension;
  }
  this->hmap = new float[dimension*dimension];
}



/*!
  Funkce uvolni pamet, ktera byla alokovana pro vyskovou mapu.
 */

void tmpTerrainPatch::freeHmap()
{
  delete[] this->hmap;
  this->hmap = NULL;
}

/*!
  Funkce aplikuje na krajinu filter.
 */
void tmpTerrainPatch::filter(float q)
{
  int x, y;

  // dky zleva doprava
  for (y = 0; y < dimension; y++)
    for (x = 1; x < dimension; x++)
      hmap[getHmapIndex(x,y)] = hmap[getHmapIndex(x-1,y)] * (1.f-q) + hmap[getHmapIndex(x,y)] * q;

  // dky zprava doleva
  for (y = 0; y < dimension; y++)
    for (x = dimension-1; x >= 1; x--)
      hmap[getHmapIndex(x,y)] = hmap[getHmapIndex(x-1,y)] * (1.f-q) + hmap[getHmapIndex(x,y)] * q;

  // sloupce zespoda nahoru
  for (y = 1; y < dimension; y++)
    for (x = 0; x < dimension; x++)
      hmap[getHmapIndex(x,y)] = hmap[getHmapIndex(x,y-1)] * (1.f-q) + hmap[getHmapIndex(x,y)] * q;

  // sloupce shora dol
  for (y = dimension-1; y >= 1; y--)
    for (x = 0; x < dimension; x++)
      hmap[getHmapIndex(x,y)] = hmap[getHmapIndex(x,y-1)] * (1.f-q) + hmap[getHmapIndex(x,y)] * q;
}

/*!
  Funkce vraci index do vyskove mapy
 */
inline int tmpTerrainPatch::getHmapIndex(const int x, const int y)
{
  return y*dimension + x;
}

/*!
  Konstruktor
 */
tmpTerrainPatch::tmpTerrainPatch()
{
  hmap = NULL;
	x1 = -1;
	y1 = -1;
	x2 = -1;
	y2 = -1;
}

/*!
  Destruktor
 */
tmpTerrainPatch::~tmpTerrainPatch()
{
	reset();
}

/*!
  Funkce vraci hodnotu z vyskove mapy
 */
float tmpTerrainPatch::getHmapValue(const int x, const int y)
{
	if ((x >= 0) && (x < dimension) && (y >= 0) && (y < dimension))
		return hmap[y*dimension + x];
	else
		return 0;
}

/*!
  Funkce vygeneruje krajinu s defaulnimi hodnotami
 */
void tmpTerrainPatch::generateHeightMap()
{
  generateHeightMap(0,0,DEFAULT_TERRAIN_DIMENSION+1,DEFAULT_TERRAIN_DIMENSION+1,
		DEFAULT_TERRAIN_FILTER_Q);
}

/*!
  Funkce vygeneruje krajinu s vlastnimi hodnotami
 */
void tmpTerrainPatch::generateHeightMap(int _x1, int _y1, int _x2, int _y2, float filterQ)
{
	//Naplnime promenne rozmeru
	x1 = _x1;
	y1 = _y1;
	x2 = _x2;
	y2 = _y2;


  // Alokujeme vkovou mapu
  allocHmap(x2-x1);

	// Vygenerujeme vyskovou mapu pomoci Perlinovy funkce
	int x,y;
	float pom;

  for (y=0; y<dimension; y++) {
	  for (x=0; x<dimension; x++) {
			pom = 150*PerlinNoise(float(x)/50,float(y)/50);
			if (pom > maxHeight) maxHeight = pom;
			if (pom < minHeight) minHeight = pom;
      hmap[getHmapIndex(x, y)] = pom;
	  }
		already_finished++;
		printf("\rHotovo: (%d%%)",already_finished*100/x2-x1);
  }

	//Vyfiltrovani krajiny
	filter(filterQ);

}

/*!
  Funkce znovunastavi krajinu.
 */
void tmpTerrainPatch::reset()
{
	freeHmap();
}

/*****************************************************************************************************************\
 *
 *      TERRAINPATCH
 *
\*****************************************************************************************************************/
/*!
  Funkce otestuje, zda je scena spravne zobrazena.
*/
bool TerrainPatch::test(float cameraX, float cameraY, float cameraZ)
{
	bool tmp_output;

	double midx = double(x1) + (x2 - x1) / 2;
	double midy = double(y1) + (y2 - y1) / 2;
	double midz = getHmapValue(dimension/2,dimension/2);
	double deltax = midx - cameraX;
	double deltay = midy - cameraY;
	double deltaz = midz - cameraZ;
	double distance2D = sqrt(deltax*deltax+deltay*deltay);
	double distance = sqrt(distance2D*distance2D + deltaz*deltaz);

	if ((((chunk_error * LODdistance) / distance) <= 1.0) || ((x2-x1) <= dimension))
		{ //chyba neni velka nebo jsme na konci stromu => tento segment ma byt vykreslen
				//zapiseme do LODtabulky
				myLODtable->setTableValue(x1/(dimension-1),y1/(dimension-1) ,x2/(dimension-1),y2/(dimension-1),LOD);
			if (model_showed == false)
			{
				model_showed = true;
				return false;
			}
			else
			{
				return true;
			}
		}
	else
		{ //pokracujeme do dalsich podsegmentu
			tmp_output = this->subs0->test(cameraX,cameraY,cameraZ);
			tmp_output = this->subs1->test(cameraX,cameraY,cameraZ) && tmp_output;
			tmp_output = this->subs2->test(cameraX,cameraY,cameraZ) && tmp_output;
			tmp_output = this->subs3->test(cameraX,cameraY,cameraZ) && tmp_output;
			if (model_showed == true)
			{
				model_showed = false;
				return false;
			}
			else
			{
				return true && tmp_output;
			}
		}
}

/*!
  Funkce prida ctverec krajiny do celkove krajiny.
*/
void TerrainPatch::generate(SoSeparator *root, float cameraX, float cameraY, float cameraZ)
{
	if (model_showed == true)
	{
		addModel(root);
	}
	else
	{
		subs0->generate(root,cameraX,cameraY,cameraZ);
		subs1->generate(root,cameraX,cameraY,cameraZ);
		subs2->generate(root,cameraX,cameraY,cameraZ);
		subs3->generate(root,cameraX,cameraY,cameraZ);
	}
}

float TerrainPatch::getHmapValue(const int x, const int y)
{
	int step = (x2 - x1) / (dimension - 1);
	return tmpTerrain->getHmapValue(x1 + x*step,y1 + y*step);
}


/*!
	Tato funkce naplni vyskovou mapu aktualniho segmentu. Data prebira z pomocneho segmentu.
*/
void TerrainPatch::generateHeightMap(int _x1, int _y1, int _x2, int _y2, int seg_dimension, int hq, int _LOD)
{
	//Naplnime LOD
	LOD = _LOD;
	//Naplnime promenne rozmeru
	x1 = _x1;
	y1 = _y1;
	x2 = _x2;
	y2 = _y2;
	dimension = seg_dimension;
	chunk_error = float(x2 - x1) / (dimension - 1);

	// Vygenerujeme poduzly
	if ((x2 - x1) <= dimension)
	{ //uz jsme na nejnizsi urovni

	}
	else
	{ //musime vygenerovat poduzly
		allocSubs();
		int midx = (x2 - x1) / 2;
		int midy = (y2 - y1) / 2;
		_LOD++;
		subs0->generateHeightMap(x1,y1,x1+midx,y1+midy,seg_dimension,hq,_LOD);
		subs1->generateHeightMap(x1+midx,y1,x2,y1+midy,seg_dimension,hq,_LOD);
		subs2->generateHeightMap(x1,y1+midy,x1+midx,y2,seg_dimension,hq,_LOD);
		subs3->generateHeightMap(x1+midx,y1+midy,x2,y2,seg_dimension,hq,_LOD);
	}


	//Vygenerovani modelu
	generateModel(float(x2-x1),hq);

	generateLeftBorder(float(x2-x1));
	generateRightBorder(float(x2-x1));
	generateTopBorder(float(x2-x1));
	generateBottomBorder(float(x2-x1));

	generateModifiedBorder(float(x2-x1),1,0);
	generateModifiedBorder(float(x2-x1),1,1);
	generateModifiedBorder(float(x2-x1),1,2);
	generateModifiedBorder(float(x2-x1),1,3);



	//Ukazatel hotovych segmentu
	already_finished++;
	printf("\rHotovo: %d (%d%%)",already_finished,already_finished*100/total_segments);
}




/*!
  Funkce vytvori zobrazitelny graf sceny modelu krajiny
  zalozene na vyskove mape. Parametr size udava velikost krajiny 
	(v prostorovych souradnicich)
  a parametr hq udava metodu vytvoreni modelu krajiny.
  0 znaci standardni tesselaci, tj. dva trojuhelniky na ctverec krajiny,
  1 je podobna predchozi volbe, ale zlom mezi temito trojuhelnky je
  vytvaren podle toho, ve kterem smeru je v dane dlazdici vetsi zmena vysky,
  2 dlazdice je rendrovana ctyrmi trojuhelniky.
*/
void TerrainPatch::generateModel(float size, int hq)
{
	//zmenime hodnoty, aby se vygenerovala pouze vnitrni cast segmentu
	int step = (x2 - x1) / (dimension - 1);
	x1 += step;
	y1 += step;
	x2 -= step;
	y2 -= step;
	size -= (2*step);
	dimension -= 2;


	// Materil (barva) krajiny,
	// rzn indexy odpovdaj rznm barvm ternu.
	SoMaterial *material = new SoMaterial;
	material->ambientColor.set1Value(0, SbColor(0.0f, 0.0f, 0.8f));
	material->diffuseColor.set1Value(0, SbColor(0.0f, 0.0f, 0.8f));
	material->ambientColor.set1Value(1, SbColor(0.0f, 0.8f, 0.0f));
	material->diffuseColor.set1Value(1, SbColor(0.0f, 0.8f, 0.0f));
	material->ambientColor.set1Value(2, SbColor(0.6f, 0.6f, 0.0f));
	material->diffuseColor.set1Value(2, SbColor(0.6f, 0.6f, 0.0f));
	material->ambientColor.set1Value(3, SbColor(0.8f, 0.8f, 0.8f));
	material->diffuseColor.set1Value(3, SbColor(0.8f, 0.8f, 0.8f));
	model->addChild(material);

  // Pro normly nastavujeme binding, tj. zda-li specifikujeme
  // normly pro kad bod (PER_VERTEX), nebo jen pro trojhelnky
  // (PER_FACE), i cel plocha m jen jednu normlu (OVERALL).
  // Ppona _INDEXED znamen, e se normly nebudou brt sekvenn
  // z nodu SoNormal, nbr se se budou indexovat ze
  // SoIndexedTriangleStripSet::normalIndex podobn jako se
  // souadnice indexuj z SoIndexedTriangleStripSet::coordIndex.
  SoNormalBinding *nbind = new SoNormalBinding;
  nbind->value.setValue(SoNormalBinding::PER_VERTEX_INDEXED);
  model->addChild(nbind);

  // Binding pro materily,
  // vce informac u pedchozho komente k SoNormalBinding.
  SoMaterialBinding *mbind = new SoMaterialBinding;
  mbind->value.setValue(SoMaterialBinding::PER_VERTEX_INDEXED);
  model->addChild(mbind);

  // Vytvome nody a nastavme poet souadnic a norml, kter
  // budeme generovat. (template SbSqr pot druhou mocninu)
  SoCoordinate3 *coords = new SoCoordinate3;
  SoNormal *normals = new SoNormal;
  int num = (hq != 2) ? SbSqr(dimension) : SbSqr(dimension)+SbSqr(dimension-1);
  coords->point.setNum(num);
  normals->vector.setNum(num);
  SbVec3f *c = coords->point.startEditing();
  SbVec3f *ctemp = c;
  SbVec3f *n = normals->vector.startEditing();


  // Nyn generujeme trojhelnky do pamti. Kad tverec z na
  // vkov mapy renderujeme jako tyi trojhelnky. Teoreticky
  // by staily dva, ale model by pi osvtlovn vykazoval chyby.
  int xi,yi;
  float sx = size / (dimension-1);
  float sy = sx;
  SbVec3f midp;

  if (hq != 2) {
    for (yi=0; yi<dimension; yi++) {
      for (xi=0; xi<dimension; xi++) {

        // souadnice a normla pro lev spodn roh tverce v krajin
        *(c++) = SbVec3f(float(x1) + xi*sx, getHmapValue(xi,yi), float(y1) + yi*sy);
        *(n++) = calculateNormal(xi, yi, sx, sy);

      }
    }
  } else {
    for (yi=0; yi<dimension-1; yi++) {
      for (xi=0; xi<dimension-1; xi++) {

        // Vpoet prostednho bodu tverce
        midp.setValue(float(x1) + (xi+0.5f) * sx,
                      (getHmapValue(xi,yi) + getHmapValue(xi+1,yi) +
                      getHmapValue(xi,yi+1)+ getHmapValue(xi+1,yi+1)) / 4.f,
                      float(y1) + (yi+0.5f) * sy);

        // Spotme si normly pro tyi rohov body i pro
        // stedov bod
        SbVec3f n1 = calculateNormal(xi, yi, sx, sy);
        SbVec3f n2 = calculateNormal(xi, yi+1, sx, sy);
        SbVec3f n3 = calculateNormal(xi+1, yi+1, sx, sy);
        SbVec3f n4 = calculateNormal(xi+1, yi, sx, sy);
        SbVec3f nmid = n1+n2+n3+n4;
        nmid.normalize();

        // souadnice a normla pro lev spodn roh tverce v krajin
        *(c++) = SbVec3f(float(x1) + xi*sx, getHmapValue(xi,yi), float(y1) + yi*sy);
        *(n++) = n1;

        // souadnice a normla pro prostedn bod ve tverci krajiny
        *(c++) = midp;
        *(n++) = nmid;
      }
      // posledn bod v dku
      *(c++) = SbVec3f(float(x1) + xi*sx, getHmapValue(xi,yi), float(y1) + yi*sy);
      *(n++) = calculateNormal(xi, yi, sx, sy);
    }

    // Posledn ada vertex
    for (xi=0; xi<dimension; xi++) {
      *(c++) = SbVec3f(float(x1) + xi*sx, getHmapValue(xi,yi), float(y1) + yi*sy);
      *(n++) = calculateNormal(xi, yi, sx, sy);
    }
  }


  // Nakonfiguruj triangleStripSet
  SoIndexedTriangleStripSet *stripSet = new SoIndexedTriangleStripSet;
  switch (hq) {
  case 0: num = dimension * (dimension-1) * 2 + dimension-1; break;
  case 1: num = dimension * (dimension-1) * 5; break;
  case 2: num = (dimension-1) * (dimension-1) * 10; break;
  }
  stripSet->coordIndex.setNum(num);
  stripSet->materialIndex.setNum(num);

  // Prvnmu ze strip vdy pia 5 souadnic (vytvo 3 trojhelnky)
  // a druhmu 3 (vytvo jeden trojhelnk).
  int32_t *ci = stripSet->coordIndex.startEditing();
  int32_t *mi = stripSet->materialIndex.startEditing();
  switch (hq) {
  case 0: {
    c = ctemp;
    int i1 = 0;
    int i2 = dimension;
    for (yi=0; yi<dimension-1; yi++) {
      for (xi=0; xi<dimension; xi++) {
        *(ci++) = i2;   // lev horn bod tverce
        *(ci++) = i1;   // lev spodn bod tverce
        // Materilov indexy pro vertexy, kter jsme vygenerovali
        // na pedchozch dcch.
        *(mi++) = getMaterialIndex(c[i2][1], minHeight, maxHeight);   // lev horn bod tverce
        *(mi++) = getMaterialIndex(c[i1][1], minHeight, maxHeight);   // lev spodn bod tverce
        i1++;
        i2++;
      }
      *(ci++) = -1;     // konec triangle stripu
      *(mi++) = -1;
    }
    break;
  }
  case 1: {
    c = ctemp;
    int i1 = 0;
    int i2 = dimension;
    for (yi=0; yi<dimension-1; yi++) {
      for (xi=0; xi<dimension-1; xi++) {
        // Potej zmnu v jednotlivch smrech dladice,
        // abychom zjistili, ve kterm smru bude dladice rozdlena.
        bool orderChange = fabsf(c[i2+1][1]-c[i1][1]) > fabsf(c[i2][1]-c[i1+1][1]);
        if (orderChange) {
          int tmp = i2;
          i2 = i1;
          i1 = tmp;
        }

        *(ci++) = i2;   // lev horn bod tverce
        *(ci++) = i1;   // lev spodn bod tverce
        // Materilov indexy pro vertexy, kter jsme vygenerovali
        // na pedchozch dcch.
        *(mi++) = getMaterialIndex(c[i2][1], minHeight, maxHeight);   // lev horn bod tverce
        *(mi++) = getMaterialIndex(c[i1][1], minHeight, maxHeight);   // lev spodn bod tverce
        i1++;
        i2++;
        *(ci++) = i2;   // prav horn bod tverce
        *(ci++) = i1;   // prav spodn bod tverce
        *(mi++) = getMaterialIndex(c[i2][1], minHeight, maxHeight);   // prav horn bod tverce
        *(mi++) = getMaterialIndex(c[i1][1], minHeight, maxHeight);   // prav spodn bod tverce
        *(ci++) = -1;
        *(mi++) = -1;
      }
      i1++;
      i2++;
    }
    break;
  }
  case 2: {
    c = ctemp;
    int tmp1,tmp2,tmp3;
    int i1 = 0;
    int i2 = 2*dimension-1;
    for (yi=0; yi<dimension-2; yi++) {
      for (xi=0; xi<dimension-1; xi++) {
        // Generovn souadnicovch index
        // (vdy dva triangleStripSety, prvn sloen ze t,
        // druh z jednoho trojhelnku)
        *(ci++) = i2;   // lev horn bod tverce
        *(ci++) = i1;   // lev spodn bod tverce
        *(ci++) = i1+1; // prostedn bod
        *(ci++) = i1+2; // prav spodn bod
        *(ci++) = i2+2; // prav horn bod
        *(ci++) = -1;
        *(ci++) = i2;   // lev horn bod
        *(ci++) = i1+1; // prostedn bod
        *(ci++) = i2+2; // prav horn bod
        *(ci++) = -1;

        // Materilov indexy pro vertexy, kter jsme vygenerovali
        // na pedchozch dcch.
        *(mi++) = tmp1 = getMaterialIndex(c[i2][1], minHeight, maxHeight);   // lev horn bod tverce
        *(mi++) =        getMaterialIndex(c[i1][1], minHeight, maxHeight);   // lev spodn bod tverce
        *(mi++) = tmp2 = getMaterialIndex(c[i1+1][1], minHeight, maxHeight); // prostedn bod
        *(mi++) =        getMaterialIndex(c[i1+2][1], minHeight, maxHeight); // prav spodn bod
        *(mi++) = tmp3 = getMaterialIndex(c[i2+2][1], minHeight, maxHeight); // prav horn bod
        *(mi++) = -1;
        *(mi++) = tmp1; // lev horn bod
        *(mi++) = tmp2; // prostedn bod
        *(mi++) = tmp3; // prav horn bod
        *(mi++) = -1;

        i1 += 2;
        i2 += 2;
      }
      i1++;
      i2++;
    }

    // Posledn dek
    for (xi=0; xi<dimension-1; xi++) {
      *(ci++) = i2;
      *(ci++) = i1;
      *(ci++) = i1+1;
      *(ci++) = i1+2;
      *(ci++) = i2+1; // zde zmna oproti pedchoz smyce
      *(ci++) = -1;
      *(ci++) = i2;
      *(ci++) = i1+1;
      *(ci++) = i2+1; // zde zmna
      *(ci++) = -1;

      *(mi++) = tmp1 = getMaterialIndex(c[i2][1], minHeight, maxHeight);
      *(mi++) =        getMaterialIndex(c[i1][1], minHeight, maxHeight);
      *(mi++) = tmp2 = getMaterialIndex(c[i1+1][1], minHeight, maxHeight);
      *(mi++) =        getMaterialIndex(c[i1+2][1], minHeight, maxHeight);
      *(mi++) = tmp3 = getMaterialIndex(c[i2+1][1], minHeight, maxHeight); // zde zmna
      *(mi++) = -1;
      *(mi++) = tmp1;
      *(mi++) = tmp2;
      *(mi++) = tmp3;
      *(mi++) = -1;

      i1 += 2;
      i2 += 1;
    }
    break;
  }}
  stripSet->coordIndex.finishEditing();
  stripSet->materialIndex.finishEditing();
  coords->point.finishEditing();
  normals->vector.finishEditing();
  model->addChild(normals);
  model->addChild(coords);
  model->addChild(stripSet);

	//vratime puvodni hodnoty
	x1 -= step;
	y1 -= step;
	x2 += step;
	y2 += step;
	size += 2;
	dimension += 2;

}

void TerrainPatch::generateLeftBorder(float size)
{
	//vygenerujeme normalni hranice
	// Materil (barva) krajiny,
	// rzn indexy odpovdaj rznm barvm ternu.
	SoMaterial *material = new SoMaterial;
	material->ambientColor.set1Value(0, SbColor(0.0f, 0.0f, 0.8f));
	material->diffuseColor.set1Value(0, SbColor(0.0f, 0.0f, 0.8f));
	material->ambientColor.set1Value(1, SbColor(0.0f, 0.8f, 0.0f));
	material->diffuseColor.set1Value(1, SbColor(0.0f, 0.8f, 0.0f));
	material->ambientColor.set1Value(2, SbColor(0.6f, 0.6f, 0.0f));
	material->diffuseColor.set1Value(2, SbColor(0.6f, 0.6f, 0.0f));
	material->ambientColor.set1Value(3, SbColor(0.8f, 0.8f, 0.8f));
	material->diffuseColor.set1Value(3, SbColor(0.8f, 0.8f, 0.8f));
	border0->addChild(material);

  // Pro normly nastavujeme binding, tj. zda-li specifikujeme
  // normly pro kad bod (PER_VERTEX), nebo jen pro trojhelnky
  // (PER_FACE), i cel plocha m jen jednu normlu (OVERALL).
  // Ppona _INDEXED znamen, e se normly nebudou brt sekvenn
  // z nodu SoNormal, nbr se se budou indexovat ze
  // SoIndexedTriangleStripSet::normalIndex podobn jako se
  // souadnice indexuj z SoIndexedTriangleStripSet::coordIndex.
  SoNormalBinding *nbind = new SoNormalBinding;
  nbind->value.setValue(SoNormalBinding::PER_VERTEX_INDEXED);
  border0->addChild(nbind);

  // Binding pro materily,
  // vce informac u pedchozho komente k SoNormalBinding.
  SoMaterialBinding *mbind = new SoMaterialBinding;
  mbind->value.setValue(SoMaterialBinding::PER_VERTEX_INDEXED);
  border0->addChild(mbind);


  // Vytvome nody a nastavme poet souadnic a norml, kter
  // budeme generovat. (template SbSqr pot druhou mocninu)
  SoCoordinate3 *coords = new SoCoordinate3;
  SoNormal *normals = new SoNormal;
  int num = (dimension * 2) - 2;
  coords->point.setNum(num);
  normals->vector.setNum(num);
  SbVec3f *c = coords->point.startEditing();
  SbVec3f *ctemp = c;
  SbVec3f *n = normals->vector.startEditing();

  SoIndexedTriangleStripSet *stripSet = new SoIndexedTriangleStripSet;
  stripSet->coordIndex.setNum(num+1);
  stripSet->materialIndex.setNum(num+1);

  int32_t *ci = stripSet->coordIndex.startEditing();
  int32_t *mi = stripSet->materialIndex.startEditing();

  // Nyn generujeme trojhelnky do pamti. Kad tverec z na
  // vkov mapy renderujeme jako tyi trojhelnky. Teoreticky
  // by staily dva, ale model by pi osvtlovn vykazoval chyby.
  int yi;
  float sx = size / (dimension-1);
  float sy = sx;

	int pocitadlo = 0;
    for (yi=0; yi<dimension; yi++) 
		{
      *(c++) = SbVec3f(float(x1), getHmapValue(0,yi), float(y1) + yi*sy);
      *(n++) = calculateNormal(0, yi, sx, sy);
			*(ci++) = pocitadlo;
      *(mi++) = getMaterialIndex(getHmapValue(0,yi), minHeight, maxHeight);   // lev horn bod tverce
			pocitadlo++;
			if ((yi != 0) && (yi != (dimension - 1)))
			{
	      *(c++) = SbVec3f(float(x1 + sx), getHmapValue(1,yi), float(y1) + yi*sy);
		    *(n++) = calculateNormal(1, yi, sx, sy);
				*(ci++) = pocitadlo;
				*(mi++) = getMaterialIndex(getHmapValue(1,yi), minHeight, maxHeight);   // lev horn bod tverce
				pocitadlo++;
			}
    }

    *(ci++) = -1;     // konec triangle stripu
    *(mi++) = -1;

  stripSet->coordIndex.finishEditing();
  stripSet->materialIndex.finishEditing();
  coords->point.finishEditing();
  normals->vector.finishEditing();
  border0->addChild(normals);
  border0->addChild(coords);
  border0->addChild(stripSet);

}

void TerrainPatch::generateRightBorder(float size)
{
	//vygenerujeme normalni hranice
	// Materil (barva) krajiny,
	// rzn indexy odpovdaj rznm barvm ternu.
	SoMaterial *material = new SoMaterial;
	material->ambientColor.set1Value(0, SbColor(0.0f, 0.0f, 0.8f));
	material->diffuseColor.set1Value(0, SbColor(0.0f, 0.0f, 0.8f));
	material->ambientColor.set1Value(1, SbColor(0.0f, 0.8f, 0.0f));
	material->diffuseColor.set1Value(1, SbColor(0.0f, 0.8f, 0.0f));
	material->ambientColor.set1Value(2, SbColor(0.6f, 0.6f, 0.0f));
	material->diffuseColor.set1Value(2, SbColor(0.6f, 0.6f, 0.0f));
	material->ambientColor.set1Value(3, SbColor(0.8f, 0.8f, 0.8f));
	material->diffuseColor.set1Value(3, SbColor(0.8f, 0.8f, 0.8f));
	border2->addChild(material);

  // Pro normly nastavujeme binding, tj. zda-li specifikujeme
  // normly pro kad bod (PER_VERTEX), nebo jen pro trojhelnky
  // (PER_FACE), i cel plocha m jen jednu normlu (OVERALL).
  // Ppona _INDEXED znamen, e se normly nebudou brt sekvenn
  // z nodu SoNormal, nbr se se budou indexovat ze
  // SoIndexedTriangleStripSet::normalIndex podobn jako se
  // souadnice indexuj z SoIndexedTriangleStripSet::coordIndex.
  SoNormalBinding *nbind = new SoNormalBinding;
  nbind->value.setValue(SoNormalBinding::PER_VERTEX_INDEXED);
  border2->addChild(nbind);

  // Binding pro materily,
  // vce informac u pedchozho komente k SoNormalBinding.
  SoMaterialBinding *mbind = new SoMaterialBinding;
  mbind->value.setValue(SoMaterialBinding::PER_VERTEX_INDEXED);
  border2->addChild(mbind);


  // Vytvome nody a nastavme poet souadnic a norml, kter
  // budeme generovat. (template SbSqr pot druhou mocninu)
  SoCoordinate3 *coords = new SoCoordinate3;
  SoNormal *normals = new SoNormal;
  int num = (dimension * 2) - 2;
  coords->point.setNum(num);
  normals->vector.setNum(num);
  SbVec3f *c = coords->point.startEditing();
  SbVec3f *ctemp = c;
  SbVec3f *n = normals->vector.startEditing();

  SoIndexedTriangleStripSet *stripSet = new SoIndexedTriangleStripSet;
  stripSet->coordIndex.setNum(num+1);
  stripSet->materialIndex.setNum(num+1);

  int32_t *ci = stripSet->coordIndex.startEditing();
  int32_t *mi = stripSet->materialIndex.startEditing();

  // Nyn generujeme trojhelnky do pamti. Kad tverec z na
  // vkov mapy renderujeme jako tyi trojhelnky. Teoreticky
  // by staily dva, ale model by pi osvtlovn vykazoval chyby.
  int yi;
  float sx = size / (dimension-1);
  float sy = sx;

	int pocitadlo = 0;
    for (yi=0; yi<dimension; yi++) 
		{
      *(c++) = SbVec3f(float(x2), getHmapValue(dimension-1,yi), float(y1) + yi*sy);
      *(n++) = calculateNormal(dimension-1, yi, sx, sy);
			*(ci++) = pocitadlo;
      *(mi++) = getMaterialIndex(getHmapValue(dimension-1,yi), minHeight, maxHeight);   // lev horn bod tverce
			pocitadlo++;
			if ((yi != 0) && (yi != (dimension - 1)))
			{
	      *(c++) = SbVec3f(float(x2 - sx), getHmapValue(dimension-2,yi), float(y1) + yi*sy);
		    *(n++) = calculateNormal(dimension-2, yi, sx, sy);
				*(ci++) = pocitadlo;
				*(mi++) = getMaterialIndex(getHmapValue(dimension-2,yi), minHeight, maxHeight);   // lev horn bod tverce
				pocitadlo++;
			}
    }

    *(ci++) = -1;     // konec triangle stripu
    *(mi++) = -1;

  stripSet->coordIndex.finishEditing();
  stripSet->materialIndex.finishEditing();
  coords->point.finishEditing();
  normals->vector.finishEditing();
  border2->addChild(normals);
  border2->addChild(coords);
  border2->addChild(stripSet);

}

void TerrainPatch::generateTopBorder(float size)
{
	//vygenerujeme normalni hranice
	// Materil (barva) krajiny,
	// rzn indexy odpovdaj rznm barvm ternu.
	SoMaterial *material = new SoMaterial;
	material->ambientColor.set1Value(0, SbColor(0.0f, 0.0f, 0.8f));
	material->diffuseColor.set1Value(0, SbColor(0.0f, 0.0f, 0.8f));
	material->ambientColor.set1Value(1, SbColor(0.0f, 0.8f, 0.0f));
	material->diffuseColor.set1Value(1, SbColor(0.0f, 0.8f, 0.0f));
	material->ambientColor.set1Value(2, SbColor(0.6f, 0.6f, 0.0f));
	material->diffuseColor.set1Value(2, SbColor(0.6f, 0.6f, 0.0f));
	material->ambientColor.set1Value(3, SbColor(0.8f, 0.8f, 0.8f));
	material->diffuseColor.set1Value(3, SbColor(0.8f, 0.8f, 0.8f));
	border1->addChild(material);

  // Pro normly nastavujeme binding, tj. zda-li specifikujeme
  // normly pro kad bod (PER_VERTEX), nebo jen pro trojhelnky
  // (PER_FACE), i cel plocha m jen jednu normlu (OVERALL).
  // Ppona _INDEXED znamen, e se normly nebudou brt sekvenn
  // z nodu SoNormal, nbr se se budou indexovat ze
  // SoIndexedTriangleStripSet::normalIndex podobn jako se
  // souadnice indexuj z SoIndexedTriangleStripSet::coordIndex.
  SoNormalBinding *nbind = new SoNormalBinding;
  nbind->value.setValue(SoNormalBinding::PER_VERTEX_INDEXED);
  border1->addChild(nbind);

  // Binding pro materily,
  // vce informac u pedchozho komente k SoNormalBinding.
  SoMaterialBinding *mbind = new SoMaterialBinding;
  mbind->value.setValue(SoMaterialBinding::PER_VERTEX_INDEXED);
  border1->addChild(mbind);


  // Vytvome nody a nastavme poet souadnic a norml, kter
  // budeme generovat. (template SbSqr pot druhou mocninu)
  SoCoordinate3 *coords = new SoCoordinate3;
  SoNormal *normals = new SoNormal;
  int num = (dimension * 2) - 2;
  coords->point.setNum(num);
  normals->vector.setNum(num);
  SbVec3f *c = coords->point.startEditing();
  SbVec3f *ctemp = c;
  SbVec3f *n = normals->vector.startEditing();

  SoIndexedTriangleStripSet *stripSet = new SoIndexedTriangleStripSet;
  stripSet->coordIndex.setNum(num+1);
  stripSet->materialIndex.setNum(num+1);

  int32_t *ci = stripSet->coordIndex.startEditing();
  int32_t *mi = stripSet->materialIndex.startEditing();

  // Nyn generujeme trojhelnky do pamti. Kad tverec z na
  // vkov mapy renderujeme jako tyi trojhelnky. Teoreticky
  // by staily dva, ale model by pi osvtlovn vykazoval chyby.
  int xi;
  float sx = size / (dimension-1);
  float sy = sx;

	int pocitadlo = 0;
    for (xi=0; xi<dimension; xi++) 
		{
      *(c++) = SbVec3f(float(x1) + xi*sx, getHmapValue(xi,0), float(y1));
      *(n++) = calculateNormal(xi, 0, sx, sy);
			*(ci++) = pocitadlo;
      *(mi++) = getMaterialIndex(getHmapValue(xi,0), minHeight, maxHeight);   // lev horn bod tverce
			pocitadlo++;
			if ((xi != 0) && (xi != (dimension - 1)))
			{
	      *(c++) = SbVec3f(float(x1) + xi*sx, getHmapValue(xi,1), float(y1) + sy);
		    *(n++) = calculateNormal(xi, 1, sx, sy);
				*(ci++) = pocitadlo;
				*(mi++) = getMaterialIndex(getHmapValue(xi,1), minHeight, maxHeight);   // lev horn bod tverce
				pocitadlo++;
			}
    }

    *(ci++) = -1;     // konec triangle stripu
    *(mi++) = -1;

  stripSet->coordIndex.finishEditing();
  stripSet->materialIndex.finishEditing();
  coords->point.finishEditing();
  normals->vector.finishEditing();
  border1->addChild(normals);
  border1->addChild(coords);
  border1->addChild(stripSet);

}

void TerrainPatch::generateBottomBorder(float size)
{
	//vygenerujeme normalni hranice
	// Materil (barva) krajiny,
	// rzn indexy odpovdaj rznm barvm ternu.
	SoMaterial *material = new SoMaterial;
	material->ambientColor.set1Value(0, SbColor(0.0f, 0.0f, 0.8f));
	material->diffuseColor.set1Value(0, SbColor(0.0f, 0.0f, 0.8f));
	material->ambientColor.set1Value(1, SbColor(0.0f, 0.8f, 0.0f));
	material->diffuseColor.set1Value(1, SbColor(0.0f, 0.8f, 0.0f));
	material->ambientColor.set1Value(2, SbColor(0.6f, 0.6f, 0.0f));
	material->diffuseColor.set1Value(2, SbColor(0.6f, 0.6f, 0.0f));
	material->ambientColor.set1Value(3, SbColor(0.8f, 0.8f, 0.8f));
	material->diffuseColor.set1Value(3, SbColor(0.8f, 0.8f, 0.8f));
	border3->addChild(material);

  // Pro normly nastavujeme binding, tj. zda-li specifikujeme
  // normly pro kad bod (PER_VERTEX), nebo jen pro trojhelnky
  // (PER_FACE), i cel plocha m jen jednu normlu (OVERALL).
  // Ppona _INDEXED znamen, e se normly nebudou brt sekvenn
  // z nodu SoNormal, nbr se se budou indexovat ze
  // SoIndexedTriangleStripSet::normalIndex podobn jako se
  // souadnice indexuj z SoIndexedTriangleStripSet::coordIndex.
  SoNormalBinding *nbind = new SoNormalBinding;
  nbind->value.setValue(SoNormalBinding::PER_VERTEX_INDEXED);
  border3->addChild(nbind);

  // Binding pro materily,
  // vce informac u pedchozho komente k SoNormalBinding.
  SoMaterialBinding *mbind = new SoMaterialBinding;
  mbind->value.setValue(SoMaterialBinding::PER_VERTEX_INDEXED);
  border3->addChild(mbind);


  // Vytvome nody a nastavme poet souadnic a norml, kter
  // budeme generovat. (template SbSqr pot druhou mocninu)
  SoCoordinate3 *coords = new SoCoordinate3;
  SoNormal *normals = new SoNormal;
  int num = (dimension * 2) - 2;
  coords->point.setNum(num);
  normals->vector.setNum(num);
  SbVec3f *c = coords->point.startEditing();
  SbVec3f *ctemp = c;
  SbVec3f *n = normals->vector.startEditing();

  SoIndexedTriangleStripSet *stripSet = new SoIndexedTriangleStripSet;
  stripSet->coordIndex.setNum(num+1);
  stripSet->materialIndex.setNum(num+1);

  int32_t *ci = stripSet->coordIndex.startEditing();
  int32_t *mi = stripSet->materialIndex.startEditing();

  // Nyn generujeme trojhelnky do pamti. Kad tverec z na
  // vkov mapy renderujeme jako tyi trojhelnky. Teoreticky
  // by staily dva, ale model by pi osvtlovn vykazoval chyby.
  int xi;
  float sx = size / (dimension-1);
  float sy = sx;

	int pocitadlo = 0;
    for (xi=0; xi<dimension; xi++) 
		{
      *(c++) = SbVec3f(float(x1) + xi*sx, getHmapValue(xi,dimension-1), float(y2));
      *(n++) = calculateNormal(xi, dimension-1, sx, sy);
			*(ci++) = pocitadlo;
      *(mi++) = getMaterialIndex(getHmapValue(xi,dimension-1), minHeight, maxHeight);   // lev horn bod tverce
			pocitadlo++;
			if ((xi != 0) && (xi != (dimension - 1)))
			{
	      *(c++) = SbVec3f(float(x1) + xi*sx, getHmapValue(xi,dimension-2), float(y2) - sy);
		    *(n++) = calculateNormal(xi, dimension-2, sx, sy);
				*(ci++) = pocitadlo;
				*(mi++) = getMaterialIndex(getHmapValue(xi,dimension-2), minHeight, maxHeight);   // lev horn bod tverce
				pocitadlo++;
			}
    }

    *(ci++) = -1;     // konec triangle stripu
    *(mi++) = -1;

  stripSet->coordIndex.finishEditing();
  stripSet->materialIndex.finishEditing();
  coords->point.finishEditing();
  normals->vector.finishEditing();
  border3->addChild(normals);
  border3->addChild(coords);
  border3->addChild(stripSet);
}


void TerrainPatch::generateModifiedBorder(float size, int level_diff, int border)
{
	if (border_level[border] != level_diff)
	{ //pouze pokud ma mit hranice jinou uroven detailu
	border_level[border] = level_diff; //nastavim uroven detailu

	//vygenerujeme normalni hranice
	// Materil (barva) krajiny,
	// rzn indexy odpovdaj rznm barvm ternu.
	SoMaterial *material = new SoMaterial;
	material->ambientColor.set1Value(0, SbColor(0.0f, 0.0f, 0.8f));
	material->diffuseColor.set1Value(0, SbColor(0.0f, 0.0f, 0.8f));
	material->ambientColor.set1Value(1, SbColor(0.0f, 0.8f, 0.0f));
	material->diffuseColor.set1Value(1, SbColor(0.0f, 0.8f, 0.0f));
	material->ambientColor.set1Value(2, SbColor(0.6f, 0.6f, 0.0f));
	material->diffuseColor.set1Value(2, SbColor(0.6f, 0.6f, 0.0f));
	material->ambientColor.set1Value(3, SbColor(0.8f, 0.8f, 0.8f));
	material->diffuseColor.set1Value(3, SbColor(0.8f, 0.8f, 0.8f));

  // Pro normly nastavujeme binding, tj. zda-li specifikujeme
  // normly pro kad bod (PER_VERTEX), nebo jen pro trojhelnky
  // (PER_FACE), i cel plocha m jen jednu normlu (OVERALL).
  // Ppona _INDEXED znamen, e se normly nebudou brt sekvenn
  // z nodu SoNormal, nbr se se budou indexovat ze
  // SoIndexedTriangleStripSet::normalIndex podobn jako se
  // souadnice indexuj z SoIndexedTriangleStripSet::coordIndex.
  SoNormalBinding *nbind = new SoNormalBinding;
  nbind->value.setValue(SoNormalBinding::PER_VERTEX_INDEXED);

  // Binding pro materily,
  // vce informac u pedchozho komente k SoNormalBinding.
  SoMaterialBinding *mbind = new SoMaterialBinding;
  mbind->value.setValue(SoMaterialBinding::PER_VERTEX_INDEXED);



  //vypocet 2^level_diff
	int step = 1;
	for (int i = 0; i < level_diff; i++)
	{
		step = step * 2;
	}

	// Vytvome nody a nastavme poet souadnic a norml, kter
  // budeme generovat. (template SbSqr pot druhou mocninu)
  SoCoordinate3 *coords = new SoCoordinate3;
  SoNormal *normals = new SoNormal;
  int num = 3 *  (((dimension-1)/step) + (((dimension-1)/step)-1) * step + step/2);
  int num2 = 4 * (((dimension-1)/step) + (((dimension-1)/step)-1) * step + step/2);
  coords->point.setNum(num);
  normals->vector.setNum(num);
  SbVec3f *c = coords->point.startEditing();
  SbVec3f *ctemp = c;
  SbVec3f *n = normals->vector.startEditing();

  SoIndexedTriangleStripSet *stripSet = new SoIndexedTriangleStripSet;
  stripSet->coordIndex.setNum(num2);
  stripSet->materialIndex.setNum(num2);

  int32_t *ci = stripSet->coordIndex.startEditing();
  int32_t *mi = stripSet->materialIndex.startEditing();

  // Nyn generujeme trojhelnky do pamti. Kad tverec z na
  // vkov mapy renderujeme jako tyi trojhelnky. Teoreticky
  // by staily dva, ale model by pi osvtlovn vykazoval chyby.
  int xi,yi;
  float sx = size / (dimension-1);
  float sy = sx;

	int pocitadlo = 0;

	switch (border)
	{
		case 0:
    for (yi=0; yi<dimension; yi+=step) 
		{
			
      if (yi < dimension - 1)
			{
			*(c++) = SbVec3f(float(x1), getHmapValue(0,yi), float(y1) + yi*sy);
      *(n++) = calculateNormal2(0, yi, sx, sy,step);
			*(ci++) = pocitadlo;
      *(mi++) = getMaterialIndex(getHmapValue(0,yi), minHeight, maxHeight);   // lev horn bod tverce
			pocitadlo++;
      *(c++) = SbVec3f(float(x1), getHmapValue(0,yi+step), float(y1) + (yi+step)*sy);
      *(n++) = calculateNormal2(0, yi+step, sx, sy,step);
			*(ci++) = pocitadlo;
      *(mi++) = getMaterialIndex(getHmapValue(0,yi+step), minHeight, maxHeight);   // lev horn bod tverce
			pocitadlo++;
      *(c++) = SbVec3f(float(x1) + sx, getHmapValue(1,yi+step/2), float(y1) + (yi + step/2)*sy);
      *(n++) = calculateNormal(1, yi + step/2, sx, sy);
			*(ci++) = pocitadlo;
      *(mi++) = getMaterialIndex(getHmapValue(1,yi + step/2), minHeight, maxHeight);   // lev horn bod tverce
			pocitadlo++;
		  *(ci++) = -1;     // konec triangle stripu
			*(mi++) = -1;
			}


				for (int i = yi - step / 2 + 1;i <= yi + step/2; i++)
				{
					if ((i > 1) && (i < dimension - 1))
					{
					//spolecny vrchol
					*(c++) = SbVec3f(float(x1), getHmapValue(0,yi), float(y1) + yi*sy);
					*(n++) = calculateNormal2(0, yi, sx, sy,step);
					*(ci++) = pocitadlo;
					*(mi++) = getMaterialIndex(getHmapValue(0,yi), minHeight, maxHeight);   // lev horn bod tverce
					pocitadlo++;

					*(c++) = SbVec3f(float(x1) + sx, getHmapValue(1,i), float(y1) + i*sy);
					*(n++) = calculateNormal(1, i, sx, sy);
					*(ci++) = pocitadlo;
					*(mi++) = getMaterialIndex(getHmapValue(1,i), minHeight, maxHeight);   // lev horn bod tverce
					pocitadlo++;

					*(c++) = SbVec3f(float(x1) + sx, getHmapValue(1,i-1), float(y1) + (i-1)*sy);
					*(n++) = calculateNormal(1, i-1, sx, sy);
					*(ci++) = pocitadlo;
					*(mi++) = getMaterialIndex(getHmapValue(1,i-1), minHeight, maxHeight);   // lev horn bod tverce
					pocitadlo++;
					*(ci++) = -1;     // konec triangle stripu
					*(mi++) = -1;
					}
				}
    }
		break;
		case 1:
    for (xi=0; xi<dimension; xi+=step) 
		{
			
      if (xi < dimension - 1)
			{
      *(c++) = SbVec3f(float(x1) + xi*sx, getHmapValue(xi,0), float(y1));
      *(n++) = calculateNormal2(xi, 0, sx, sy,step);
			*(ci++) = pocitadlo;
      *(mi++) = getMaterialIndex(getHmapValue(xi,0), minHeight, maxHeight);   // lev horn bod tverce
			pocitadlo++;
      *(c++) = SbVec3f(float(x1) + (xi+step)*sx, getHmapValue(xi+step,0), float(y1));
      *(n++) = calculateNormal2(xi+step, 0, sx, sy,step);
			*(ci++) = pocitadlo;
      *(mi++) = getMaterialIndex(getHmapValue(xi+step,0), minHeight, maxHeight);   // lev horn bod tverce
			pocitadlo++;
      *(c++) = SbVec3f(float(x1) + (xi + step/2)*sx, getHmapValue(xi+step/2,1), float(y1) + sy);
      *(n++) = calculateNormal(xi + step/2,1, sx, sy);
			*(ci++) = pocitadlo;
      *(mi++) = getMaterialIndex(getHmapValue(xi + step/2,1), minHeight, maxHeight);   // lev horn bod tverce
			pocitadlo++;
		  *(ci++) = -1;     // konec triangle stripu
			*(mi++) = -1;
			}


				for (int i = xi - step / 2 + 1;i <= xi + step/2; i++)
				{
					if ((i > 1) && (i < dimension - 1))
					{
					//spolecny vrchol
					*(c++) = SbVec3f(float(x1) + xi*sx, getHmapValue(xi,0), float(y1));
					*(n++) = calculateNormal2(xi, 0, sx, sy,step);
					*(ci++) = pocitadlo;
					*(mi++) = getMaterialIndex(getHmapValue(xi,0), minHeight, maxHeight);   // lev horn bod tverce
					pocitadlo++;

					*(c++) = SbVec3f(float(x1) + i*sx, getHmapValue(i,1), float(y1) + sy);
					*(n++) = calculateNormal(i, 1, sx, sy);
					*(ci++) = pocitadlo;
					*(mi++) = getMaterialIndex(getHmapValue(i,1), minHeight, maxHeight);   // lev horn bod tverce
					pocitadlo++;

					*(c++) = SbVec3f(float(x1) + (i-1)*sx, getHmapValue(i-1,1), float(y1) + sy);
					*(n++) = calculateNormal(i-1, 1, sx, sy);
					*(ci++) = pocitadlo;
					*(mi++) = getMaterialIndex(getHmapValue(i-1,1), minHeight, maxHeight);   // lev horn bod tverce
					pocitadlo++;
					*(ci++) = -1;     // konec triangle stripu
					*(mi++) = -1;
					}
				}
    }
		break;
		case 2:
    for (yi=0; yi<dimension; yi+=step) 
		{
			
      if (yi < dimension - 1)
			{
      *(c++) = SbVec3f(float(x2), getHmapValue(dimension - 1,yi), float(y1) + yi*sy);
      *(n++) = calculateNormal2(dimension - 1, yi, sx, sy,step);
			*(ci++) = pocitadlo;
      *(mi++) = getMaterialIndex(getHmapValue(dimension - 1,yi), minHeight, maxHeight);   // lev horn bod tverce
			pocitadlo++;
      *(c++) = SbVec3f(float(x2), getHmapValue(dimension - 1,yi+step), float(y1) + (yi+step)*sy);
      *(n++) = calculateNormal2(dimension - 1, yi+step, sx, sy,step);
			*(ci++) = pocitadlo;
      *(mi++) = getMaterialIndex(getHmapValue(dimension - 1,yi+step), minHeight, maxHeight);   // lev horn bod tverce
			pocitadlo++;
      *(c++) = SbVec3f(float(x2) - sx, getHmapValue(dimension - 2,yi+step/2), float(y1) + (yi + step/2)*sy);
      *(n++) = calculateNormal(dimension - 2, yi + step/2, sx, sy);
			*(ci++) = pocitadlo;
      *(mi++) = getMaterialIndex(getHmapValue(dimension - 2,yi + step/2), minHeight, maxHeight);   // lev horn bod tverce
			pocitadlo++;
		  *(ci++) = -1;     // konec triangle stripu
			*(mi++) = -1;
			}


				for (int i = yi - step / 2 + 1;i <= yi + step/2; i++)
				{
					if ((i > 1) && (i < dimension - 1))
					{
					//spolecny vrchol
					*(c++) = SbVec3f(float(x2), getHmapValue(dimension - 1,yi), float(y1) + yi*sy);
					*(n++) = calculateNormal2(dimension - 1, yi, sx, sy,step);
					*(ci++) = pocitadlo;
					*(mi++) = getMaterialIndex(getHmapValue(dimension - 1,yi), minHeight, maxHeight);   // lev horn bod tverce
					pocitadlo++;

					*(c++) = SbVec3f(float(x2) - sx, getHmapValue(dimension - 2,i), float(y1) + i*sy);
					*(n++) = calculateNormal(dimension - 2, i, sx, sy);
					*(ci++) = pocitadlo;
					*(mi++) = getMaterialIndex(getHmapValue(dimension - 2,i), minHeight, maxHeight);   // lev horn bod tverce
					pocitadlo++;

					*(c++) = SbVec3f(float(x2) - sx, getHmapValue(dimension - 2,i-1), float(y1) + (i-1)*sy);
					*(n++) = calculateNormal(dimension - 2, i-1, sx, sy);
					*(ci++) = pocitadlo;
					*(mi++) = getMaterialIndex(getHmapValue(dimension - 2,i-1), minHeight, maxHeight);   // lev horn bod tverce
					pocitadlo++;
					*(ci++) = -1;     // konec triangle stripu
					*(mi++) = -1;
					}
				}
    }
		break;
		case 3:
    for (xi=0; xi<dimension; xi+=step) 
		{
			
      if (xi < dimension - 1)
			{
      *(c++) = SbVec3f(float(x1) + xi*sx, getHmapValue(xi,dimension-1), float(y2));
      *(n++) = calculateNormal2(xi, dimension-1, sx, sy,step);
			*(ci++) = pocitadlo;
      *(mi++) = getMaterialIndex(getHmapValue(xi,dimension-1), minHeight, maxHeight);   // lev horn bod tverce
			pocitadlo++;
      *(c++) = SbVec3f(float(x1) + (xi+step)*sx, getHmapValue(xi+step,dimension-1), float(y2));
      *(n++) = calculateNormal2(xi+step, dimension-1, sx, sy,step);
			*(ci++) = pocitadlo;
      *(mi++) = getMaterialIndex(getHmapValue(xi+step,dimension-1), minHeight, maxHeight);   // lev horn bod tverce
			pocitadlo++;
      *(c++) = SbVec3f(float(x1) + (xi + step/2)*sx, getHmapValue(xi+step/2,dimension-2), float(y2) - sy);
      *(n++) = calculateNormal(xi + step/2,dimension-2, sx, sy);
			*(ci++) = pocitadlo;
      *(mi++) = getMaterialIndex(getHmapValue(xi + step/2,dimension-2), minHeight, maxHeight);   // lev horn bod tverce
			pocitadlo++;
		  *(ci++) = -1;     // konec triangle stripu
			*(mi++) = -1;
			}


				for (int i = xi - step / 2 + 1;i <= xi + step/2; i++)
				{
					if ((i > 1) && (i < dimension - 1))
					{
					//spolecny vrchol
					*(c++) = SbVec3f(float(x1) + xi*sx, getHmapValue(xi,dimension-1), float(y2));
					*(n++) = calculateNormal2(xi, dimension-1, sx, sy,step);
					*(ci++) = pocitadlo;
					*(mi++) = getMaterialIndex(getHmapValue(xi,dimension-1), minHeight, maxHeight);   // lev horn bod tverce
					pocitadlo++;

					*(c++) = SbVec3f(float(x1) + i*sx, getHmapValue(i,dimension-2), float(y2) - sy);
					*(n++) = calculateNormal(i, dimension-2, sx, sy);
					*(ci++) = pocitadlo;
					*(mi++) = getMaterialIndex(getHmapValue(i,dimension-2), minHeight, maxHeight);   // lev horn bod tverce
					pocitadlo++;

					*(c++) = SbVec3f(float(x1) + (i-1)*sx, getHmapValue(i-1,dimension-2), float(y2) - sy);
					*(n++) = calculateNormal(i-1, dimension-2, sx, sy);
					*(ci++) = pocitadlo;
					*(mi++) = getMaterialIndex(getHmapValue(i-1,dimension-2), minHeight, maxHeight);   // lev horn bod tverce
					pocitadlo++;
					*(ci++) = -1;     // konec triangle stripu
					*(mi++) = -1;
					}
				}
    }
		break;
	}

  //*(ci++) = -1;     // konec triangle stripu
  //*(mi++) = -1;

  stripSet->coordIndex.finishEditing();
  stripSet->materialIndex.finishEditing();
  coords->point.finishEditing();
  normals->vector.finishEditing();
	switch (border)
	{
		case 0:
			border0m->removeAllChildren();
			border0m->addChild(material);
		  border0m->addChild(nbind);
		  border0m->addChild(mbind);
			border0m->addChild(normals);
			border0m->addChild(coords);
			border0m->addChild(stripSet);
			break;
		case 1:
			border1m->removeAllChildren();
			border1m->addChild(material);
		  border1m->addChild(nbind);
		  border1m->addChild(mbind);
			border1m->addChild(normals);
			border1m->addChild(coords);
			border1m->addChild(stripSet);
			break;
		case 2:
			border2m->removeAllChildren();
			border2m->addChild(material);
		  border2m->addChild(nbind);
		  border2m->addChild(mbind);
			border2m->addChild(normals);
			border2m->addChild(coords);
			border2m->addChild(stripSet);
			break;
		case 3:
			border3m->removeAllChildren();
			border3m->addChild(material);
		  border3m->addChild(nbind);
		  border3m->addChild(mbind);
			border3m->addChild(normals);
			border3m->addChild(coords);
			border3m->addChild(stripSet);
			break;
	}
	} //konec - pouze pokud bude mit jinou uroven detailu
}

/*!
	Tato funkce prida model ho stromu root.
*/
void TerrainPatch::addModel(SoSeparator *root)
{
	root->addChild(model);

	//kontrola svu	
	int left = LOD - myLODtable->getTableValue((x1/(dimension-1))-1,y1/(dimension-1));
	int top = LOD - myLODtable->getTableValue(x1/(dimension-1),(y1/(dimension-1))-1);
	int right = LOD - myLODtable->getTableValue((x2/(dimension-1)),y1/(dimension-1));
	int bottom = LOD - myLODtable->getTableValue(x1/(dimension-1),(y2/(dimension-1)));
	
	
	if ((left > 0) && (left < LOD))
	{
		generateModifiedBorder(float(x2-x1),left,0);
		root->addChild(border0m);
	}
	else
	{
		root->addChild(border0);
	}
	if ((top > 0) && (top < LOD))
	{
		generateModifiedBorder(float(x2-x1),top,1);
		root->addChild(border1m);
	}
	else
	{
		root->addChild(border1);
	}
	if ((right > 0) && (right < LOD))
	{
		generateModifiedBorder(float(x2-x1),right,2);
		root->addChild(border2m);
	}
	else
	{
		root->addChild(border2);
	}
	if ((bottom > 0) && (bottom < LOD))
	{
		generateModifiedBorder(float(x2-x1),bottom,3);
		root->addChild(border3m);
	}
	else
	{
		root->addChild(border3);
	}
	
/*
	root->addChild(border0);
	root->addChild(border1);
	root->addChild(border2);
	root->addChild(border3);

	root->addChild(border0m);
	root->addChild(border1m);
	root->addChild(border2m);
	root->addChild(border3m);
*/
}

/*!
  Funkce alokuje model.
 */
void TerrainPatch::allocModel()
{
  if (this->model != NULL) {
		freeModel();
	}
	else
	{
    this->model = new SoSeparator;
	}
	model->ref();
}



/*!
  Funkce uvoln pamt, kter byla alokovna pro model.
 */
void TerrainPatch::freeModel()
{
  this->model->unref();
}

void TerrainPatch::allocBorders()
{
  if (this->border0 != NULL) {
		freeBorders();
	}
	else
	{
    this->border0 = new SoSeparator;
    this->border1 = new SoSeparator;
    this->border2 = new SoSeparator;
    this->border3 = new SoSeparator;
    this->border0m = new SoSeparator;
    this->border1m = new SoSeparator;
    this->border2m = new SoSeparator;
    this->border3m = new SoSeparator;
	}
	border0->ref();
	border1->ref();
	border2->ref();
	border3->ref();
	border0m->ref();
	border1m->ref();
	border2m->ref();
	border3m->ref();
}

void TerrainPatch::freeBorders()
{
  this->border0->unref();
  this->border1->unref();
  this->border2->unref();
  this->border3->unref();
  this->border0m->unref();
  this->border1m->unref();
  this->border2m->unref();
  this->border3m->unref();
}

/*!
  Funkce alokuje poduzly
 */
void TerrainPatch::allocSubs()
{
  delete subs0;
  delete subs1;
  delete subs2;
  delete subs3;
	subs0 = new TerrainPatch();
	subs1 = new TerrainPatch();
	subs2 = new TerrainPatch();
	subs3 = new TerrainPatch();
}

/*!
  Funkce uvolni poduzly.
 */
void TerrainPatch::freeSubs()
{
  delete subs0;
  delete subs1;
  delete subs2;
  delete subs3;
}


/*!
  Funkce vypot normlu v danm bod vkov mapy.
 */
SbVec3f TerrainPatch::calculateNormal(const int x, const int y, const float sx, const float sy)
{
  float dx,dy;
	int step = (x2 - x1) / (dimension - 1);
  // hel v ose X
  if (x>=1 && x<=dimension-2)
    // Vpoet ze dvou sousednch bod
    dx = (getHmapValue(x-1,y) - getHmapValue(x+1,y)) / 2.f;
  else
    // Vpoet na okraji mapy (pouziji se hodnoty z pomocneho segmentu pro lepsi navazovani segmentu)
    if (x<1)  dx = (tmpTerrain->getHmapValue((x-1)*step + x1,y*step + y1) - getHmapValue(x+1,y)) / 2;
    else  dx = (getHmapValue(x-1,y) - tmpTerrain->getHmapValue((x+1)*step + x1,y*step + y1)) / 2;

  // hel v ose Y
  if (y>=1 && y<=dimension-2)
    dy = (getHmapValue(x,y-1) - getHmapValue(x,y+1)) / 2.f;
  else
    // Vpoet na okraji mapy (pouziji se hodnoty z pomocneho segmentu pro lepsi navazovani segmentu)
    if (y<1)  dy = (tmpTerrain->getHmapValue(x*step + x1,(y-1)*step + y1) - getHmapValue(x,y+1)) / 2;
    else  dy = (getHmapValue(x,y-1) - tmpTerrain->getHmapValue(x*step + x1,(y+1)*step + y1)) / 2;


  SbVec3f r(dx/sx, 1.f, dy/sy);
  r.normalize();
  return r;
}

/*!
  Funkce vypot normlu v danm bod vkov mapy.
 */
SbVec3f TerrainPatch::calculateNormal2(const int x, const int y, const float sx, const float sy, int step)
{
  float dx,dy;
  // hel v ose X
  dx = (getHmapValue(x-step,y) - getHmapValue(x+step,y)) / 2.f;

  // hel v ose Y
   dy = (getHmapValue(x,y-step) - getHmapValue(x,y+step)) / 2.f;


  SbVec3f r(dx/(sx*step), 1.f, dy/(sy*step));
  r.normalize();
  return r;
}

/*!
  Funkce vrt index materilu pro dan bod vkov mapy.
 */
int TerrainPatch::getMaterialIndex(const float height, const float minHeight, const float maxHeight)
{
  float part = (maxHeight-minHeight) / 5.f;
  if (height <= minHeight+3.f*part)  return 1; // zelen materil
  if (height <= minHeight+4.f*part)  return 2; // utrovit materil
  return 3; // snh
}



/*!
  Konstruktor.
*/
TerrainPatch::TerrainPatch()
{
	subs0 = NULL;
	subs1 = NULL;
	subs2 = NULL;
	subs3 = NULL;
	model = NULL;
	border0 = NULL;
	border1 = NULL;
	border2 = NULL;
	border3 = NULL;
	border0m = NULL;
	border1m = NULL;
	border2m = NULL;
	border3m = NULL;
	border_level[0] = 0;
	border_level[1] = 0;
	border_level[2] = 0;
	border_level[3] = 0;
	x1 = -1;
	y1 = -1;
	x2 = -1;
	y2 = -1;
	model_showed = false;
	chunk_error = 0;
	allocModel();
	allocBorders();
}



/*!
  Destruktor.
*/
TerrainPatch::~TerrainPatch()
{
  reset();
}



/*!
  Inicializuje objekt do podobnho stavu jako pi vytvoen
  objektu, tzn. minimln vyuit zdroj.
  Funkce se pouv pedevm pro uvolnn pamti
  vkov mapy, nen-li ji dle potebn.
*/
void TerrainPatch::reset()
{
	freeSubs();
	freeModel();
	freeBorders();
}



// Funkce pouit v sablonov funkci getParam pro konverzi
// string do rznch typ.
void strtoX(int &r, const char *nptr, char **endptr)
{
  r = strtol(nptr, endptr, 10);
}
void strtoX(float &r, const char *nptr, char **endptr)
{
  r = strtof(nptr, endptr);
}
void strtoX(bool &r, const char *nptr, char **endptr)
{
  int i;
  strtoX(i, nptr, endptr);
  r = (i != 0);
}

// Sablonov funkce pro ten parametr rznch typ z pkazov dky
template<class Type>
bool getParam(int argc, char **argv, int &position,
              const char *param, Type &result)
{
  if (argc > position && strcmp(argv[position], param) == 0) {
    position++;
    if (argc > position) {
      char *errp;
      Type tmp;
      strtoX(tmp, argv[position], &errp);
      if (*(argv[position]) != '\0' && *errp == '\0') {
        result = tmp;
        position++;
        return true;
      } else {
#if !defined(__WIN32__) && !defined(_WIN32)
        printf("Chybn argument \"%s\".", param);
#else
        printf("Chybny argument \"%s\".", param);
#endif
        exit(EXIT_FAILURE);
      }
    } else {
#if !defined(__WIN32__) && !defined(_WIN32)
      printf("Chybjc hodnota pro parametr \"%s\".", param);
#else
      printf("Chybejici hodnota pro parametr \"%s\".", param);
#endif
      exit(EXIT_FAILURE);
    }
  }
  return false;
}


// Funkce vytvo graf scny pro nasi aplikaci.
SoSeparator* createScene(int argc, char **argv)
{
  int i = 1;

  // Vpis npovdy
  if (argc > 1 && (strcmp(argv[1], "-h") == 0 || strcmp(argv[1], "--help") == 0)) {
    printf("Pouziti: %s [-h|--help] [-seed int] [-dimension int] [-size float]\n"
           "         [-heightQ float] [-filterQ float] [-hq int]\n"
           "-h, --help        tato obrazovka\n"
           "-seed int         nastaveni generatoru nahodnych cisel\n"
           "-size int         velikost krajiny\n"
           "-dimension int    velikost segmentu\n"
           "-distance float   vzdalenost od ktere se snizuje LOD\n"
           "-filterQ float    parametr filtrovaci funkce\n"
           "-hq int           hodnota 0,1 a 2 meni rezim tesselace krajiny\n"
           "                  (0 standard, 1 adaptivni smer hrany, 2 ctyri trojuhelniky)\n"
           "Priklad pouziti (se standardnimi hodnotami parametru krome seed):\n"
           "%s -seed 10 -size 512 -dimension 33 -distance 100.0 -filterQ 0.5 -hq 0\n",
           argv[0], argv[0]);
    exit(EXIT_SUCCESS);
  }

  // Inicializace genertoru nhodnch sel
  int seed;
  if (getParam<int>(argc, argv, i, "-seed", seed))
    srand(seed);
  else
    srand((unsigned int)time(NULL));

	//Inicializuje Perlinovu funkci (konstanty)
	PerlinInit();

  // Vytvo koen scny
  SoSeparator *root = new SoSeparator;
  root->ref();

  // Nastav intenzitu globlnho svtla ve scn
  SoEnvironment *env = new SoEnvironment;
  env->ambientIntensity.setValue(0.5);
  root->addChild(env);

  // Svtlo
  SoDirectionalLight *sunLight = new SoDirectionalLight;
  sunLight->direction.setValue(0.35f, -1.0f, 0.25f);
  sunLight->intensity.setValue(1.0f);
  root->addChild(sunLight);

  // Model krajiny
  int size = DEFAULT_TERRAIN_DIMENSION;
  int dimension = DEFAULT_SEGMENT_DIMENSION;
  float filterQ = DEFAULT_TERRAIN_FILTER_Q;
  int hq = DEFAULT_TERRAIN_HQ;
	LODdistance = float(sqrt(3)*dimension);
  getParam<int>(argc, argv, i, "-size", size);
  getParam<int>(argc, argv, i, "-dimension", dimension);
  getParam<float>(argc, argv, i, "-distance", LODdistance);
  getParam<float>(argc, argv, i, "-filterQ", filterQ);
  getParam<int>(argc, argv, i, "-hq", hq);

  // Zkontroluj, zda byly zparsovny spn vechny argumenty z command-line.
  if (i != argc) {
#if !defined(__WIN32__) && !defined(_WIN32)
    printf("Chybn parametr %s.\n", argv[i]);
#else
    printf("Chybny parametr %s.\n", argv[i]);
#endif
    exit(EXIT_FAILURE);
  }

  // Zkontroluj platnost parametr danch uivatelem.
  // size mus bt mocnina 2
  if (int(pow(2, int(log(size)/log(2.f)+0.5f))+0.5f) != size) {
#if !defined(__WIN32__) && !defined(_WIN32)
    printf("Parametr size mus bt mocnina dvou.\n"
           "Napklad 256,512,1024,2048 a podobn.\n");
#else
    printf("Parametr size musi byt mocnina dvou.\n"
           "Napriklad 256,512,1024,2048 a podobne.\n");
#endif
    exit(EXIT_FAILURE);
  }
  // Zkontroluj platnost parametr danch uivatelem.
  // dimension mus bt mocnina 2 plus 1
  if (int(pow(2, int(log(dimension-1)/log(2.f)+0.5f))+0.5f) != dimension-1) {
#if !defined(__WIN32__) && !defined(_WIN32)
    printf("Parametr dimension mus bt mocnina dvou plus 1.\n"
           "Napklad 17,33,65,129 a podobn.\n");
#else
    printf("Parametr dimension musi byt mocnina dvou plus 1.\n"
           "Napriklad 17,33,65,129 a podobne.\n");
#endif
    exit(EXIT_FAILURE);
  }
  // filterQ mus bt v rozsahu od 0.f do 1.f
  if (filterQ < 0.f || filterQ > 1.f) {
#if !defined(__WIN32__) && !defined(_WIN32)
    printf("Parametr filterQ mus bt v rozsahu 0.0 a 1.0.\n"
           "Pklad: 0.5, 0.36 a podobn.\n");
#else
    printf("Parametr filterQ musi byt v rozsahu 0.0 az 1.0.\n"
           "Priklad: 0.5, 0.36 a podobne.\n");
#endif
    exit(EXIT_FAILURE);
  }
  // Omez hq na hodnoty 0,1 a 2
  if (hq < 0 || hq > 2)  hq = 2;

  // Generovn krajiny
  printf("Generuji krajinu Perlinovou funkci:\n"
         "velikost terenu:    %i\n"
         "velikost segmentu:  %i\n"
         "LOD vzdalenost:     %.2f\n"
         "filterQ:            %.2f\n"
         "hq:                 %i\n",
         size,dimension, LODdistance, filterQ, hq); 
	
	//Vygenerovani krajiny do pomocneho segmentu
	tmpTerrain = new tmpTerrainPatch();
	already_finished = 0;
	printf("Generovani vyskove mapy:\n");
	tmpTerrain->generateHeightMap(0,0,size+1,size+1,filterQ);

	//priprava LODtabulky
	myLODtable = new LODtable();
	myLODtable->allocTable(size/(dimension-1));

	//Vygenerovani Quad-Tree
	Terrain = new TerrainPatch();
	for (int ii = size/(dimension-1);ii >= 1;ii=ii/2)
	{
		total_segments += ((ii) * (ii));
	}

	already_finished = 0;
	printf("\nGenerovani segmentu krajiny:\n");
	Terrain->generateHeightMap(0,0,size,size,dimension,hq,1);


	//Prvotni vygenerovani modelu
	Terrain->test(float(size) / 2,float(size),0);
	Terrain->generate(root,float(size) / 2,float(size),0);
  root->unrefNoDelete();
  return root;
}

void LODtable::allocTable(int dimension)
{
  if (this->table != NULL) {
    delete[] this->table;
  }
	this->dimension = dimension;
  this->table = new int[dimension*dimension];
}

void LODtable::freeTable()
{
  delete[] this->table;
  this->table = NULL;
}

inline int LODtable::getTableIndex(const int x, const int y)
{
	return y*dimension + x;
}

int LODtable::getTableValue(const int x, const int y)
{
	if ((x >= 0) && (y >= 0) && (x < dimension) && (y < dimension))
		return table[this->getTableIndex(x,y)];
	else
		return 0;
}

void LODtable::setTableValue(int x1, int y1, int x2, int y2, int value)
{
	for (int y = y1; y < y2; y++)
	{
		for (int x = x1; x < x2; x++)
		{
			table[this->getTableIndex(x,y)] = value;
		}
	}
}

LODtable::LODtable()
{
	table = NULL;
	dimension = 0;
}

LODtable::~LODtable()
{
	freeTable();
}

