/*****************************************************************************\
 *
 * perlin_f.cpp
 *
 * perlin_f - implementation of 2D Perlin noise
 *
 * Authors: Petr "Lamik" Vrba (xvrbap00 _AT stud.fit.vutbr.cz)
 *
 * ----------------------------------------------------------------------------
 *
 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 * This source code is offered for use in the public domain.
 * You may use, modify or distribute it freely.
 *
 * This source code is distributed in the hope that it will be useful but
 * WITHOUT ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY
 * DISCLAIMED.  This includes but is not limited to warranties of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * If you find the source code useful, authors will kindly welcome
 * if you give them credit and keep their names with their source code,
 * but do not feel to be forced to do so.
 *
\*****************************************************************************/

#include <stdlib.h>
#include "perlin_f.h"

int k1 = 15731;
int k2 = 789221;
int k3 = 1376312589;

/*!
	Funkce vracejici nahodnou int hodnotu v rozsahu 0 az max - 1
*/
int randi(int max)
{
	return (int((rand() * max)));
}

/*!
	Inicializuje Perlinovy funkci (konstanty)
*/
void PerlinInit()
{
	k1 = k1 + randi(2500);
	k2 = k2 + randi(10000);
	k3 = k3 + randi(100000000);
}

inline float Interpolate(float a, float b, float x)
{
	return  a*(1-x) + b*x;
}

/*!
	Vraci hrubou hodnotu jedne iterace Perlinovy funkce
*/
float Noise(int x, int y)
{
	int n;
	n = x + y * 57;
	n = (n<<13) ^ n;

	return float( 1.0 - float( (n * (n * n * k1 + k2) + k3) & 0x7fffffff) / 1073741824.0);    
}

/*!
	Vraci vyhlazenou hodnotu jedne iterace Perlinovy funkce
*/
float SmoothNoise(int x, int y)
{
	float corners, sides, center;
	corners = ( Noise(x-1, y-1) + Noise(x+1, y-1) + Noise(x-1, y+1) + Noise(x+1, y+1) ) / 16;
	sides   = ( Noise(x-1, y)  +Noise(x+1, y)  +Noise(x, y-1)  +Noise(x, y+1) ) /  8;
	center  =  Noise(x, y) / 4;
	return corners + sides + center;
}

/*!
	Vraci interpolovanou hodnotu jedne iterace Perlinovy funkce
*/
float InterpolatedNoise(float x, float y)
{
	int integer_X    = int(x);
	int integer_Y    = int(y);

	float fractional_X = x - integer_X;

	float fractional_Y = y - integer_Y;

	float v1,v2,v3,v4,i1,i2;

	v1 = SmoothNoise(integer_X,     integer_Y);
	v2 = SmoothNoise(integer_X + 1, integer_Y);
	v3 = SmoothNoise(integer_X,     integer_Y + 1);
	v4 = SmoothNoise(integer_X + 1, integer_Y + 1);

	i1 = Interpolate(v1 , v2 , fractional_X);
	i2 = Interpolate(v3 , v4 , fractional_X);

	return Interpolate(i1 , i2 , fractional_Y);
}

/*!
	Vraci kompletni hodnotu Perlinovy funkce
*/
float PerlinNoise(float x, float y)
{
      float total = 0;
      float p = 0.5; //persistence
      int n = 5; //pocet iteraci
			int i;

			for(i = 0; i <= n; i++)
			{
				int j;
				int frequency = 1;
				float amplitude = p;

				for (j = 0; j < i; j++) frequency *= 2;
				for (j = 0; j < i; j++) amplitude *= p;

				//total = total + Noise(x * frequency, y * frequency) * amplitude;
				//total = total + SmoothNoise(x * frequency, y * frequency) * amplitude;
				total = total + InterpolatedNoise(x * frequency, y * frequency) * amplitude;
			}

			return total;
}