/*****************************************************************************\
 *
 * chunkedLOD.h
 *
 * chunkedLOD - header for chunkedLOD class
 *
 * Authors: Petr "Lamik" Vrba (xvrbap00 _AT stud.fit.vutbr.cz)
 *
 * ----------------------------------------------------------------------------
 *
 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 * This source code is offered for use in the public domain.
 * You may use, modify or distribute it freely.
 *
 * This source code is distributed in the hope that it will be useful but
 * WITHOUT ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY
 * DISCLAIMED.  This includes but is not limited to warranties of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * If you find the source code useful, authors will kindly welcome
 * if you give them credit and keep their names with their source code,
 * but do not feel to be forced to do so.
 *
\*****************************************************************************/ 

#ifndef _CHUNKEDLOD_H
#define _CHUNKEDLOD_H

#include <Inventor/SoPrimitiveVertex.h>
#include <Inventor/SbBSPTree.h>
#include <Inventor/actions/SoCallbackAction.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoIndexedFaceSet.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoTextureCoordinate2.h>
#include <Inventor/fields/SoMFFloat.h>
#include <Inventor/fields/SoMFInt32.h>

#define LOD_DISTANCE 2.1f
#define DEFAULT_LEVEL 3
#define DEFAULT_REDUCE 50

class chunkedLOD
{
private:
	SoSeparator * root;
	SoSeparator * originalModel;
	SoCoordinate3 * coord3;
	SbBSPTree * coord3_tmp;
	SoIndexedFaceSet * ifs;
	SoTexture2 * texture2;
	SoTextureCoordinate2 * tex_coord2;

	int segmentsDone;
	int totalSegments;

	//! callback function for load all data from model
	static void triangle_cb(void * userdata, SoCallbackAction * action, const SoPrimitiveVertex * v1, const SoPrimitiveVertex * v2, const SoPrimitiveVertex * v3);

	//! callback function for load texture from input file
	static SoCallbackAction::Response texture_cb(void * userdata, SoCallbackAction * action, const SoNode * tex);
		
	//! quicksort for sorting triangles to elemination
	void QuickSort(SoMFFloat * tmp_distance, SoMFInt32 * tmp_delete, SoMFInt32 * tmp_join, int left, int right);

public:
	//! constructor
	chunkedLOD();

	//! destructor
	~chunkedLOD();

	//! return the original model
	inline SoSeparator * getOriginalModel()
	{
		return originalModel;
	}

	//! load model from iv file
	bool LoadModel(const char * filename);

	//! save model to iv file
	void SaveModel(SoSeparator * model, const char * filename);

	//! find minimal and maximal values from model
	void FindMinMax(SoSeparator * origin, float &minx, float &miny, float &maxx, float &maxy);

	//! add texture to model
	void AddTexture(SoSeparator * origin);

	//! find and eliminate bad triangles
	void CorrectTriangles(SoCoordinate3 * tmp_coord, SoIndexedFaceSet * tmp_ifs);

	//! cut segment (NON rectangular cutting)
	SoSeparator* CutSegment(SoSeparator *origin, float x1, float y1, float x2, float y2, SbBSPTree * border);

	//! reduce number of triangles
	void SimplifySegment(SoSeparator *origin, SbBSPTree * border, int reduce);

	//! create one LOD segment
	SoSeparator * CreateLevel(float x1, float y1, float x2, float y2, int level, int maxLevel, int reduce);

	//! build chunkedLOD quadtree
	SoSeparator * BuildChunkedLOD(int maxLevel, int reduce);

};
#endif